package jp.co.sds_corp.tool.test;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import jp.co.sds_corp.util.config.ConfigUtil;
import jp.co.sds_corp.util.config.DOMConfiguration;
import jp.co.sds_corp.util.config.IConfigurable;
import jp.co.sds_corp.util.config.IConfiguration;
import jp.co.sds_corp.util.diag.MemoryDiagnostics;
import jp.co.sds_corp.util.diag.SystemDiagnostics;
import jp.co.sds_corp.util.diag.ThreadMonitor;
import jp.co.sds_corp.util.log.CmnLogger;

public class TestCommand implements StressExecutor<IConfiguration>
{
	static CmnLogger logger = CmnLogger.getInstance();
	TestRunner runner;
	IConfiguration config;
	List<IConfiguration> listCase;
	Map<IConfiguration,TestRunner> mapRunner = new HashMap<IConfiguration,TestRunner>();
	StressTester<IConfiguration> stress;
	int poolSize = 10;
	int saveHist = 200;
	int saveStart = 1;
	int history = 1;
	List<String> listHist = new ArrayList<String>();
	String lastCmd = null;
	String testCase;
	long interval;
	int count;
	boolean showCommandLine = true;
	Properties prop = new Properties();
	SystemDiagnostics sysDiag = new SystemDiagnostics();
	MemoryDiagnostics memDiag = new MemoryDiagnostics();
	ThreadMonitor threadMon = new ThreadMonitor();
	ShutdownHook shut = new ShutdownHook();

	TestCommand() {
		Runtime.getRuntime().addShutdownHook( shut );
	}

	public TestCommand(TestRunner runner) {
		this();
		this.runner = runner;
	}

	class ShutdownHook extends Thread {
		public void run() {
			try {
				stopService();
			} catch (Exception ex) {
				logger.error(this, "stop error", ex);
			}
		}
	}

	public void setPoolSize(int size) {
		this.poolSize = size;
	}

	public void init() throws Exception {
		config = DOMConfiguration.findResource( "testcase.xml" );
		init(config);
	}

	public void init(IConfiguration config) throws Exception {
		this.config = config;

		testCase = ConfigUtil.getProperty(config, "testCase");
		if (testCase != null) {
			showCommandLine = false;
		}
		Long lvar = ConfigUtil.getLongProperty(config, "interval");
		if (lvar != null) {
			interval = lvar.longValue();
		} else if (!showCommandLine) {
			throw(new Exception("interval not defined."));
		}
		Integer ivar = ConfigUtil.getIntProperty(config, "count");
		if (ivar != null) {
			count = ivar.intValue();
		} else if (!showCommandLine) {
			throw(new Exception("count not defined."));
		}

		initCase(config);

		stress = new StressTester<IConfiguration>( this );
		String var;
		var = System.getProperty("jp.co.smbc.aidw.tool.test.TestCommand.saveHist");
		if (var != null && var.length() > 0) {
			try {
				saveHist = Integer.valueOf(var).intValue();
			} catch (Exception ex) {
			}
		}
		var = System.getProperty("jp.co.smbc.aidw.tool.test.TestCommand.poolSize");
		if (var != null && var.length() > 0) {
			try {
				poolSize = Integer.valueOf(var).intValue();
			} catch (Exception ex) {
			}
		}
		stress.setPoolSize(poolSize);
	}

	void initCase(IConfiguration config) throws Exception {
		if (runner != null) {
			if (runner instanceof IConfigurable) {
				((IConfigurable )runner).init(config);
			}
		}
		listCase = config.listChild( "case" );
		for (IConfiguration caseCfg : listCase) {
			String cls = caseCfg.getAttributes().get("class", null);
			if (cls != null && cls.length() > 0) {
				Class clazz = Class.forName(cls);
				TestRunner runner = (TestRunner )clazz.newInstance();
				if (runner instanceof IConfigurable) {
					((IConfigurable )runner).init(caseCfg);
				}
				mapRunner.put(caseCfg, runner);
			}
		}
	}

	void startService() throws Exception {
		if (runner != null) {
			runner.startService();
		}
		for (TestRunner runner : mapRunner.values()) {
			runner.startService();
		}
	}

	public void start() throws Exception {
		if (!showCommandLine) {
			background();
		}
		BufferedReader rd = new BufferedReader(
								new InputStreamReader( System.in ));
		for ( ;; ) {
			System.out.print( "test:" + history + "> " );
			String cmd = rd.readLine();
			if ( cmd == null ) {
				break;
			}
			if ( cmd.length() == 0 ) {
				continue;
			}
			if ( cmd.charAt( cmd.length() - 1 ) == '\n' ) {
				System.out.println( "\\n" );
				cmd = cmd.substring( 0, cmd.length() - 1 );
			}

			try {
				if ( command( cmd )) {
					break;
				}
			} catch ( Exception ex ) {
				ex.printStackTrace();
			}
		}
		rd.close();
	}

	void background() throws Exception {
		List<IConfiguration> list = parseNo( testCase );
		stress.test( list, interval, count );
	}

	boolean command( String input ) throws Exception {
		List<Command> listCmd = Command.parse(input);
		listCmd = replaceHistory(listCmd);
		if (listCmd == null) {
			return false;
		}

		String save = concatCommand(listCmd);
		if (isReplaced(listCmd)) {
			System.out.println(save);
		}
		saveCommand(save);

		for (Command cmd : listCmd) {
			String command = cmd.getCommand();
			if ( command.equals( "quit" )) {
				return true;
			} else
			if ( command.equals( "start" )) {
				startService();
			} else
			if ( command.equals( "stop" )) {
				stopService();
			} else
			if ( command.equals( "run" )) {
				testService();
			} else
			if ( command.equals( "list" )) {
				list();
			} else
			if ( command.startsWith( "test " )) {
				test( command );
			} else
			if ( command.startsWith( "break" )) {
				stress.breakTest();
			} else
			if ( command.startsWith( "show " )) {
				show( command );
			} else
			if ( command.startsWith( "set " )) {
				setProperty( command );
			} else
			if ( command.equals( "history" )) {
				showHistory();
			} else
			if ( command.equals( "help" )) {
				showHelp();
			} else {
				System.out.println( "? " + command );
			}
		}

		return false;
	}

	List<Command> replaceHistory(List<Command> listCmd) {
		List<Command> ret = new ArrayList<Command>();
		for (Command cmd : listCmd) {
			String command = cmd.getCommand();
			String replace = null;
			if ( command.charAt(0) == '!') {
				replace = searchHistory(command);
				if (replace == null) {
					System.out.println( "history not found: " + command );
					return null;
				}
				List<Command> listRep = Command.parse(replace);
				for (Command repCmd : listRep) {
					repCmd.setReplaced(true);
					ret.add(repCmd);
				}
			} else {
				ret.add(cmd);
			}
		}
		return ret;
	}

	String searchHistory(String cmd) {
		String command = null;
		if (cmd.length() < 2) {
			return null;
		}
		String query = cmd.substring(1);
		if (cmd.charAt(1) == '!') {
			if (listHist.size() > 0) {
				command = listHist.get(listHist.size() - 1);
			}
		} else {
			try {
				int hist = Integer.valueOf(query);
				command = listHist.get(hist - saveStart);
			} catch (Exception ex) {
			}
		}
		if (command == null) {
			for (int i = listHist.size(); i > 0; i--) {
				String saved = listHist.get(i - 1);
				if (saved.startsWith(query)) {
					command = saved;
					break;
				}
			}
		}
		return command;
	}

	boolean isReplaced(List<Command> listCmd) {
		boolean replaced = false;
		for (Command cmd : listCmd) {
			if (cmd.isReplaced()) {
				replaced = true;
				break;
			}
		}
		return replaced;
	}

	String concatCommand(List<Command> listCmd) {
		StringBuilder save = new StringBuilder();
		for (int i = 0; i < listCmd.size(); i++ ) {
			if (i > 0) {
				save.append(" ; ");
			}
			Command cmd = listCmd.get(i);
			save.append(cmd.getCommand());
		}
		return save.toString();
	}

	void saveCommand(String cmd) {
		history++;
		listHist.add(cmd);
		if (listHist.size() > saveHist) {
			listHist.remove(0);
			saveStart++;
		}
	}

	void list() throws Exception {
		DecimalFormat df = new DecimalFormat( "##0" );
		for ( int i = 0; i < listCase.size(); i++ ) {
			IConfiguration config = listCase.get( i );
			TestRunner runner = mapRunner.get(config);
			String msg = "\t" + df.format( i ) + ") ";
			if (runner == null) {
				msg += "test=" + config.getAttributes().get( "test", "" )
					+ " type=" + config.getAttributes().get( "type", "" )
					+ " " + config.getAttributes().get( "description", "" );
			} else {
				msg += "runner=" + runner.toString();
			}
			System.out.println( msg );
		}
	}

	void show(String cmd) {
		List<String> list = null;
		String prm = cmd.substring(5);
		if (prm.equals("response")) {
			list = stress.getResponse();
		} else
		if (prm.equals("system")) {
			list = sysDiag.getSystemProperties();
		} else
		if (prm.equals("properties")) {
			showProperties();
		} else
		if (prm.equals("memory")) {
			list = memDiag.getMemoryInfo();
		} else
		if (prm.equals("thread")) {
			threadMon.setShowStack(false);
			list = threadMon.listThread();
		} else
		if (prm.equals("stack")) {
			threadMon.setShowStack(true);
			list = threadMon.listThread();
		} else {
			System.out.println( "unknown parameter[" + prm + "]." );
			return;
		}
		if (list != null) {
			for (String msg : list) {
				System.out.println(msg);
			}
		}
	}

	void showProperties() {
		Iterator itr = prop.keySet().iterator();
		while (itr.hasNext()) {
			String key = (String )itr.next();
			String value = prop.getProperty(key);
			System.out.println( "\t" + key + "=" + value );
		}
	}

	void setProperty(String cmd) {
		String prm = cmd.substring(4);
		String[] arr = prm.split("=");
		if (arr.length !=2) {
			System.out.println( "illegal format[" + prm + "]." );
			return;
		}
		String key = arr[0];
		String value = arr[1];
		prop.put(key, value);
	}

	void testService() throws Exception {
		for ( IConfiguration config : listCase ) {
			testCase( config, null );
		}
	}

	void test( String cmd ) throws Exception {
		String param = cmd.substring( 5 );
		long interval = 100;
		String[] arrPrm = param.split( " " );
		List<IConfiguration> list = parseNo( arrPrm[0] );

		switch( arrPrm.length ) {
		case 1:
			for ( IConfiguration config : list ) {
				testCase( config, null );
			}
			break;
		case 2:
			for ( IConfiguration config : list ) {
				testCase( config, arrPrm[1] );
			}
			break;
		case 3:
		default:
			int count = 1;
			try {
				interval = Long.valueOf( arrPrm[1] ).longValue();
			} catch ( Exception ex ) {
				System.err.println( "invalid interval [" + arrPrm[1] + "]" );
				return;
			}
			try {
				count = Integer.valueOf( arrPrm[2] ).intValue();
			} catch ( Exception ex ) {
				System.err.println( "invalid count [" + arrPrm[2] + "]" );
				return;
			}
			boolean waitResult = true;
			String var = prop.getProperty("wait");
			if (var != null && var.equals("false")) {
				waitResult = false;
			}
			stress.setWaitResult(waitResult);
			stress.test( list, interval, count );
			break;
		}
	}

	void testCase(IConfiguration config, String param) throws Exception {
		TestRunner runner = mapRunner.get(config);
		if (runner == null) {
			runner = this.runner;
		}
		if (runner == null) {
			logger.error(this, "TestRunner not found.");
			return;
		}
		logger.debug(this, "start test");
		runner.testCase( config, param );
	}

	@Override
	public void startThread() throws Exception {
		if (runner != null) {
			runner.startThread();
		}
		for (TestRunner runner : mapRunner.values()) {
			runner.startThread();
		}
	}

	@Override
	public void test( IConfiguration config ) throws Exception {
		testCase( config, null );
	}

	@Override
	public void stopThread() throws Exception {
		if (runner != null) {
			runner.stopThread();
		}
		for (TestRunner runner : mapRunner.values()) {
			runner.stopThread();
		}
	}

	List<IConfiguration> parseNo( String param ) throws Exception {
		List<IConfiguration> ret = new ArrayList<IConfiguration>();
		IConfiguration config;
		int idx;
		String[] arr = param.split( "," );
		for ( int i = 0; i < arr.length; i++ ) {
			if ( arr[i].indexOf( "-" ) < 0 ) {
				idx = Integer.valueOf( arr[i] ).intValue();
				config = listCase.get( idx );
				ret.add( config );
			} else {
				String[] se = arr[i].split( "-" );
				int start, end;
				try {
					start = Integer.valueOf( se[0] ).intValue();
				} catch ( Exception ex ) {
					start = 0;
				}
				if ( se.length == 1 ) {
					end = listCase.size() - 1;
				} else {
					end = Integer.valueOf( se[1] ).intValue();
				}
				for ( int j = start; j <= end; j++ ) {
					config = listCase.get( j );
					ret.add( config );
				}
			}
		}
		return ret;
	}

	void showHelp() {
		System.out.println( "quit\t\t: quit runner" );
		System.out.println( "start\t\t: start service." );
		System.out.println( "stop\t\t: stop service." );
		System.out.println( "run\t\t: run all tests." );
		System.out.println( "list\t\t: list case" );
		System.out.println( "test #\t\t: execute test. ex.test 0,1,2-5" );
		System.out.println( "test # [interval] [count]" );
		System.out.println( "\t\t: execute stress test. ex.test 0,1,2-3 10 1000" );
		System.out.println( "history\t\t: show history" );
		System.out.println( "!#\t\t: execute history #." );
		System.out.println( "!!\t\t: execute recent history." );
		System.out.println( "show response\t: show response" );
		System.out.println( "show system\t: show system properties" );
		System.out.println( "show memory\t: show memory" );
		System.out.println( "show thread\t: show thread list" );
		System.out.println( "show stack\t: show thread with stack trace" );
		System.out.println( "set [key]=[value]: set properties" );
		System.out.println( "\twait\t: wait test result [true|false]" );
		System.out.println( "help\t\t: show this help" );
	}

	void showHistory() {
		for (int i = 0; i < listHist.size(); i++) {
			String hist = listHist.get(i);
			System.out.println( String.format("%4d", (i + saveStart))
				+ "  " + hist);
		}
	}

	void usage() {
		System.out.println( "usage: " + getClass().getName() + " [TestRunner]" );
	}

	void stopService() throws Exception {
		if (stress != null) {
			stress.stop();
		}
		if (runner != null) {
			runner.stopService();
		}
		for (TestRunner runner : mapRunner.values()) {
			runner.stopService();
		}
	}

	public static void main( String[] args ) {
		TestCommand obj = new TestCommand();
		if (args.length > 0) {
			try {
				Class clazz = Class.forName( args[0] );
				obj.runner = (TestRunner )clazz.newInstance();
			} catch ( Exception ex ) {
				logger.error( TestCommand.class, "load error", ex );
				obj.usage();
				return;
			}
		}
		try {
			obj.init();
			obj.startService();
			obj.start();
			obj.stopService();
		} catch ( Exception ex ) {
			logger.error( TestCommand.class, "test error", ex );
			System.exit(9);
		}
		System.exit(0);
	}

}
